/* 
 *  Arnold emulator (c) Copyright, Kevin Thacker 1995-2001
 *  
 *  This file is part of the Arnold emulator source code distribution.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/********************/
/* snapshot support */
/********************/

/* These snapshots are in CPCEMU compatible form */
/* - ink values are stored with bits 5,6,7 = 0 */

#include <stdio.h>
#include <stdlib.h>
#include "snapshot.h"
#include "z80/z80.h"
#include "cpc.h"
#include <memory.h>
#include "host.h"
#include <string.h>
#include "snapv3.h"

extern char *Z80MemoryBase;


BOOL    Snapshot_Load(char *SnapshotFilename)
{
        unsigned long SnapshotLength;
        unsigned char   *pSnapshot;
        SNAPSHOT_HEADER *pSnapshotHeader;

        /* attempt to load snapshot data */
        Host_LoadFile(SnapshotFilename, &pSnapshot, &SnapshotLength);

        /* failed so quit */
        if (pSnapshot!=NULL)
        {
                pSnapshotHeader = (SNAPSHOT_HEADER *)pSnapshot;

                /* snapshot length must be at least the size of snapshot header */
                if (SnapshotLength>sizeof(SNAPSHOT_HEADER))
                {
                        /* is snapshot header text present? */
                        if (memcmp(pSnapshotHeader, SNAPSHOT_HEADER_TEXT, 8)==0)
                        {
                                /* header present */

                                /* check version */
                                if ((pSnapshotHeader->Version==1) || (pSnapshotHeader->Version==2) || (pSnapshotHeader->Version==3))
                                {
                                        char    *pSnapshotData;
                                        int i;

                                        /* get memory dump size */
                                        int MemSizeK = (pSnapshotHeader->MemSizeLow & 0x0ff) | ((pSnapshotHeader->MemSizeHigh & 0x0ff)<<8);
                                        int MemSizeBytes = MemSizeK<<10;

                                        /* file must be at least the size of the header + the number of bytes
                                        it claims to contain */
                                        if (SnapshotLength>=(sizeof(SNAPSHOT_HEADER)+MemSizeBytes))
                                        {
                                                /* Appears snapshot is valid. Setup Z80 and copy memory */

											int CPCType;
											int Register;

                                                /**** INIT Z80 ****/
                                                /* setup Z80 */
                                                char *pRegs = (char *)pSnapshotHeader + 0x011;

                                                /* reset CPC */
                                                CPC_Reset();

												/* get CPC type */
												CPCType = ((char *)pSnapshotHeader)[0x06d];
                                               

												if (
													(pSnapshotHeader->Version==1) ||
													(pSnapshotHeader->Version==2))
												{

													/* version 1 and 2 did not know about CPC+ so do not allow them
													to be set */
													
													/* set CPC type accordingly */
													switch (CPCType)
													{
														case 0:
														{
															CPC_SetCPCType(CPC_TYPE_CPC464);
														}
														break;

														case 1:
														{
															CPC_SetCPCType(CPC_TYPE_CPC664);
														}
														break;

														default:
														case 2:
														{
															CPC_SetCPCType(CPC_TYPE_CPC6128);
														}
														break;

													}
												}
												else
												if (pSnapshotHeader->Version==3)
												{
													/* version 3 understands CPC+ */
													/* allow them to be set */

													/* set CPC type accordingly */
													switch (CPCType)
													{
														case 0:
														{
															CPC_SetCPCType(CPC_TYPE_CPC464);
														}
														break;

														case 1:
														{
															CPC_SetCPCType(CPC_TYPE_CPC664);
														}
														break;
														case 2:
														{
															CPC_SetCPCType(CPC_TYPE_CPC6128);
														}
														break;


														/* 6 is GX4000 */
														case 6:
														case 4:
														{
															CPC_SetCPCType(CPC_TYPE_464PLUS);
														}
														break;

														default:
														case 5:
														{
															CPC_SetCPCType(CPC_TYPE_6128PLUS);
														}
														break;
													}
												}

												Register = (pRegs[1]<<8) | (pRegs[0]);
												Z80_SetReg(Z80_AF, Register);

												Register = (pRegs[3]<<8) | (pRegs[2]);
												Z80_SetReg(Z80_BC, Register);

												Register = (pRegs[5]<<8) | (pRegs[4]);
												Z80_SetReg(Z80_DE, Register);
												
												Register = (pRegs[7]<<8) | (pRegs[6]);
												Z80_SetReg(Z80_HL, Register);

												Z80_SetReg(Z80_R, pRegs[8]);

												Z80_SetReg(Z80_I, pRegs[9]);

                                                Z80_SetReg(Z80_IFF1, pRegs[10] & 1);

                                                Z80_SetReg(Z80_IFF2, pRegs[11] & 1);

												Register = (pRegs[13]<<8) | (pRegs[12]);
												Z80_SetReg(Z80_IX, Register);

												Register = (pRegs[15]<<8) | (pRegs[14]);
												Z80_SetReg(Z80_IY, Register);
								
												Register = (pRegs[17]<<8) | (pRegs[16]);
												Z80_SetReg(Z80_SP, Register);
												
												Register = (pRegs[19]<<8) | (pRegs[18]);
												Z80_SetReg(Z80_PC, Register);

												Register = pRegs[20] & 3;

												if (Register==3)
												{
													Register = 0;
												}

												Z80_SetReg(Z80_IM, Register);

												Register = (pRegs[22]<<8) | (pRegs[21]);
												Z80_SetReg(Z80_AF2, Register);

												Register = (pRegs[24]<<8) | (pRegs[23]);
												Z80_SetReg(Z80_BC2, Register);

												Register = (pRegs[26]<<8) | (pRegs[25]);
												Z80_SetReg(Z80_DE2, Register);

												Register = (pRegs[28]<<8) | (pRegs[27]);
												Z80_SetReg(Z80_HL2, Register);


                                                /**** GATE ARRAY ****/
                                                /* initialise colour palette */
                                                for (i=0; i<17; i++)
                                                {
                                                        unsigned char HwColourIndex = ((char *)pSnapshotHeader)[0x02f + i];

                                                        /* pen select */
                                                        GateArray_Write(i);
                                                
                                                        /* write colour for pen */
                                                        GateArray_Write((HwColourIndex & 0x01f) | 0x040);
                                                }

                                                /* pen select */
                                                GateArray_Write(((char *)pSnapshotHeader)[0x02e] & 0x01f);
                                                /* mode and rom configuration select */
                                                GateArray_Write((((char *)pSnapshotHeader)[0x040] & 0x03f) | 0x080);
                                                /* ram configuration select */
                                                GateArray_Write((((char *)pSnapshotHeader)[0x041] & 0x03f) | 0x0c0);

                                                /**** CRTC ****/
                                                /* initialise CRTC */
                                                for (i=0; i<18; i++)
                                                {
                                                        unsigned char CRTCRegData = ((char *)pSnapshotHeader)[0x043 + i];

                                                        CRTC_RegisterSelect(i);

                                                        CRTC_WriteData(CRTCRegData & 0x0ff);
                                                }

                                                /* select CRTC register */
                                                CRTC_RegisterSelect(((char *)pSnapshotHeader)[0x042]);

                                                /**** ROM INDEX ****/
                                                /* select expansion rom */
                                                ROM_Select(((char *)pSnapshotHeader)[0x055]);

                                                /**** PIO ****/
                                                /* set control */
                                                PPI_WriteControl(((char *)pSnapshotHeader)[0x059]);

                                                /* set port A data */
                                                PPI_SetPortAData(((char *)pSnapshotHeader)[0x056]);
                                                /* set port B data */
                                                PPI_SetPortBData(((char *)pSnapshotHeader)[0x057]);
                                                /* set port C data */
                                                PPI_SetPortCData(((char *)pSnapshotHeader)[0x058]);

                                                /**** PSG ****/
                                                /* setup PSG registers */
                                                for (i=0; i<16; i++)
                                                {
                                                        unsigned char PSGRegData = ((char *)pSnapshotHeader)[0x05b + i];

                                                        PSG_RegisterSelect(i);

                                                        PSG_WriteData(PSGRegData);
                                                }

                                                /* select PSG register */
                                                PSG_RegisterSelect(((char *)pSnapshotHeader)[0x05a]);

												if (pSnapshotHeader->Version==3)
												{
													int CRTCType;

													CRTCType = ((char *)pSnapshotHeader)[0x0a4];

													if ((CRTCType<0) || (CRTCType>=5))
													{
														CRTCType = 0;
													}

													CRTC_SetType(CRTCType);
												}


                                                /**** INIT RAM ****/
                                                /* setup RAM */
                                                pSnapshotData = ((char *)pSnapshotHeader + sizeof(SNAPSHOT_HEADER));

                                                for (i=0; i<MemSizeBytes; i++)
                                                {
                                                        Z80MemoryBase[i] = pSnapshotData[i];
                                                }

                                                /* free memory that held snapshot */
                                                free(pSnapshot);


												if (pSnapshotHeader->Version ==3)
												{
													unsigned long SizeRemaining;

													RIFF_CHUNK *pChunk = (RIFF_CHUNK *)((char *)pSnapshotHeader + sizeof(SNAPSHOT_HEADER)+MemSizeBytes);

													SizeRemaining = SnapshotLength - MemSizeBytes - sizeof(SNAPSHOT_HEADER);

													while (SizeRemaining!=0)
													{
														int ChunkSize;

														ChunkSize = Riff_GetChunkLength(pChunk);

														SnapshotV3_HandleChunk(pChunk);

														pChunk = (RIFF_CHUNK *)ChunkSize;
													
														SizeRemaining -= ChunkSize;
													}
												}

                                                return TRUE;
                                        }

                                }

                        }
                }

                /* free loaded data */
                free(pSnapshot);
        }
        
        return FALSE;
}

BOOL    Snapshot_Save(char *SnapshotFilename, int SnapshotSizeInK)
{
	const int Version = 2;
        char    *pSnapshot;
        int             SnapshotMemorySize;
        int             SnapshotSize;

        /* initially only write 64K snapshots */
        SnapshotMemorySize = SnapshotSizeInK*1024;

        /* calculate size of output file */
        SnapshotSize = SnapshotMemorySize + sizeof(SNAPSHOT_HEADER);

        /* allocate memory for header and data */
        pSnapshot = (char *)malloc(SnapshotSize);

        if (pSnapshot!=NULL)
        {
                int i;
                char *pSnapshotData;
                SNAPSHOT_HEADER *pSnapshotHeader = (SNAPSHOT_HEADER *)pSnapshot;
				int CPCType;

                /* clear header */
                memset(pSnapshotHeader, 0, sizeof(SNAPSHOT_HEADER));

                /* setup header */
                memcpy(pSnapshotHeader, SNAPSHOT_HEADER_TEXT, 8);

                /* set version */
                pSnapshotHeader->Version = (char)Version;

				/* convert my CPC type index to Snapshot CPC type index.
				They may be the same at the moment, but may change later */
				switch (CPC_GetCPCType())
				{
					case CPC_TYPE_CPC464:
					{
						CPCType = 0;
					}
					break;

					case CPC_TYPE_CPC664:
					{
						CPCType = 1;
					}
					break;

					case CPC_TYPE_CPC6128:
					{
						CPCType = 2;
					}
					break;

					case CPC_TYPE_464PLUS:
					{
						CPCType = 3;
					}
					break;

					case CPC_TYPE_6128PLUS:
					{
						CPCType = 4;
					}
					break;

					case CPC_TYPE_KCCOMPACT:
					{
						CPCType = 5;
					}
					break;
		
					default:
					{
						CPCType = 2;
					}


				}

                /* get cpc type */
                pSnapshotHeader->CPCType = (char)CPCType;

                /* ***** Z80 CPU ***** */
                {
						int Register;

                        char *pRegs = (char *)pSnapshotHeader + 0x011;
#if 0
                        pZ80Reg = Z80_GetReg();

                        pRegs[0] = pZ80Reg->AF.B.l;
                        pRegs[1] = pZ80Reg->AF.B.h;
                        pRegs[2] = pZ80Reg->BC.B.l;
                        pRegs[3] = pZ80Reg->BC.B.h;
                        pRegs[4] = pZ80Reg->DE.B.l;
                        pRegs[5] = pZ80Reg->DE.B.h;
                        pRegs[6] = pZ80Reg->HL.B.l;
                        pRegs[7] = pZ80Reg->HL.B.h;
                        pRegs[8] = (char)((pZ80Reg->R & 0x07f) | pZ80Reg->RBit7);
                        pRegs[9] = pZ80Reg->I;
                        
                        /* iff0 */
                        /* iff1 */

                        if (pZ80Reg->IFF1 == 1)
                        {
                                pRegs[10] = 1;
                        }
                        else
                        {
                                pRegs[10] = 0;
                        }

                        if (pZ80Reg->IFF2 == 1)
                        {
                                pRegs[11] = 1;
                        }
                        else
                        {
                                pRegs[11] = 0;
                        }

                        pRegs[12] = pZ80Reg->IX.B.l;
                        pRegs[13] = pZ80Reg->IX.B.h;
                        pRegs[14] = pZ80Reg->IY.B.l;
                        pRegs[15] = pZ80Reg->IY.B.h;
                        pRegs[16] = pZ80Reg->SP.B.l;
                        pRegs[17] = pZ80Reg->SP.B.h;
                        pRegs[18] = pZ80Reg->PC.B.l;
                        pRegs[19] = pZ80Reg->PC.B.h;
                        pRegs[20] = (char)(pZ80Reg->IM & 0x03);
                        pRegs[21] = pZ80Reg->altAF.B.l;
                        pRegs[22] = pZ80Reg->altAF.B.h;
                        pRegs[23] = pZ80Reg->altBC.B.l;
                        pRegs[24] = pZ80Reg->altBC.B.h;
                        pRegs[25] = pZ80Reg->altDE.B.l;
                        pRegs[26] = pZ80Reg->altDE.B.h;
                        pRegs[27] = pZ80Reg->altHL.B.l;
                        pRegs[28] = pZ80Reg->altHL.B.h;
#endif
                
						Register = Z80_GetReg(Z80_AF);

						pRegs[0] = (char)Register;
						pRegs[1] = (char)(Register>>8);

						Register = Z80_GetReg(Z80_BC);

						pRegs[2] = (char)Register;
						pRegs[3] = (char)(Register>>8);

						Register = Z80_GetReg(Z80_DE);

						pRegs[4] = (char)Register;
						pRegs[5] = (char)(Register>>8);

						Register = Z80_GetReg(Z80_HL);

						pRegs[6] = (char)Register;
						pRegs[7] = (char)(Register>>8);

						pRegs[8] = (char)Z80_GetReg(Z80_R);
						pRegs[9] = (char)Z80_GetReg(Z80_I);

						 /* iff0 */
                        /* iff1 */

                        if (Z80_GetReg(Z80_IFF1)!=0)
						{
							pRegs[10] = 1;
						}
						else
						{
							pRegs[10] = 0;
						}
						
                        if (Z80_GetReg(Z80_IFF2)!=0)
						{
                                pRegs[11] = 1;
                        }
                        else
                        {
                                pRegs[11] = 0;
                        }
				
						Register = Z80_GetReg(Z80_IX);

						pRegs[12] = (char)Register;
						pRegs[13] = (char)(Register>>8);

						Register = Z80_GetReg(Z80_IY);

						pRegs[14] = (char)Register;
						pRegs[15] = (char)(Register>>8);

						Register = Z80_GetReg(Z80_SP);

						pRegs[16] = (char)Register;
						pRegs[17] = (char)(Register>>8);

						Register = Z80_GetReg(Z80_PC);
						pRegs[18] = (char)Register;
						pRegs[19] = (char)(Register>>8);

						Register = Z80_GetReg(Z80_IM);
						pRegs[20] = (char)(Register & 0x03);
						
						Register = Z80_GetReg(Z80_AF2);
						pRegs[21] = (char)Register;
						pRegs[22] = (char)(Register>>8);

						Register = Z80_GetReg(Z80_BC2);
						pRegs[23] = (char)Register;
						pRegs[24] = (char)(Register>>8);

						Register = Z80_GetReg(Z80_DE2);
						pRegs[25] = (char)Register;
						pRegs[26] = (char)(Register>>8);

						Register = Z80_GetReg(Z80_HL2);
						pRegs[27] = (char)Register;
						pRegs[28] = (char)(Register>>8);
               	}

                /**** GATE ARRAY ****/
                /* store colour palette */
                for (i=0; i<17; i++)
                {
                   ((char *)pSnapshotHeader)[0x02f + i] = (char)((GateArray_GetPaletteColour(i) & 0x01f));
                }

                /* store selected pen */
                ((char *)pSnapshotHeader)[0x02e] = (char)(GateArray_GetSelectedPen());

                /* store mode and rom configuration */
                ((char *)pSnapshotHeader)[0x040] = (char)(GateArray_GetModeAndRomConfiguration());

                /* store ram configuration */
                ((char *)pSnapshotHeader)[0x041] = (char)(GateArray_GetRamConfiguration());

                /**** CRTC ****/
                /* get CRTC status */
                for (i=0; i<18; i++)
                {
                        ((char *)pSnapshotHeader)[0x043 + i] = CRTC_GetRegisterData(i);
                }

                /* select CRTC register */
                ((char *)pSnapshotHeader)[0x042] = (char)CRTC_GetSelectedRegister();

                /**** ROM INDEX ****/
                /* select expansion rom */
                ((char *)pSnapshotHeader)[0x055] = (char)ROM_GetSelectedROM();


                /**** PSG ****/
                /* get PSG status */
                for (i=0; i<16; i++)
                {
                        ((char *)pSnapshotHeader)[0x05b + i] = (char)PSG_GetRegisterData(i);
                }

                /* select PSG register */
                ((char *)pSnapshotHeader)[0x05a] = (char)PSG_GetSelectedRegister();
                
                /**** PPI ****/
                /* get PPI control */
                ((char *)pSnapshotHeader)[0x059] = (char)PPI_GetControl();

                /* get port A data */
                ((char *)pSnapshotHeader)[0x056] = (char)PPI_GetPortAData();
                /* get port B data */
                ((char *)pSnapshotHeader)[0x057] = (char)PPI_GetPortBData();
                /* get port C data */
                ((char *)pSnapshotHeader)[0x058] = (char)PPI_GetPortCData();


                /* set memory size */
                pSnapshotHeader->MemSizeLow = (char)(SnapshotMemorySize>>10); 
                pSnapshotHeader->MemSizeHigh = (char)((SnapshotMemorySize>>10)>>8);

                pSnapshotData = (char *)((char *)pSnapshotHeader + sizeof(SNAPSHOT_HEADER));

                for (i=0; i<SnapshotMemorySize; i++)
                {
                        pSnapshotData[i] = Z80MemoryBase[i];
                }

                /* write file */
                Host_SaveFile(SnapshotFilename,(unsigned char *)pSnapshot,SnapshotSize);
                
                /* free data */
                free(pSnapshot);
        }

        return TRUE;
}

